#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import sys
import time
import logging
from queue import Queue, Empty
from pc_ble_driver_py.observers import *
from pc_ble_driver_py.exceptions import NordicSemiException

TARGET_DEV_NAME = "hivo"
CONNECTIONS = 1
CFG_TAG = 1


def init(conn_ic_id):
    # noinspection PyGlobalUndefined
    global config, BLEDriver, BLEAdvData, BLEEvtID, BLEAdapter, BLEEnableParams, BLEGapTimeoutSrc, BLEUUID, BLEConfigCommon, BLEConfig, BLEConfigConnGatt, BLEGapScanParams
    from pc_ble_driver_py import config

    config.__conn_ic_id__ = conn_ic_id
    # noinspection PyUnresolvedReferences
    from pc_ble_driver_py.ble_driver import (
        BLEDriver,
        BLEAdvData,
        BLEEvtID,
        BLEEnableParams,
        BLEGapTimeoutSrc,
        BLEUUID,
        BLEGapScanParams,
        BLEConfigCommon,
        BLEConfig,
        BLEConfigConnGatt,
    )

    # noinspection PyUnresolvedReferences
    from pc_ble_driver_py.ble_adapter import BLEAdapter

    global nrf_sd_ble_api_ver
    nrf_sd_ble_api_ver = config.sd_api_ver_get()


class NusClient(BLEDriverObserver, BLEAdapterObserver):
    def __init__(self, adapter):
        super(NusClient, self).__init__()
        self.adapter = adapter
        self.conn_q = Queue()
        self.adapter.observer_register(self)
        self.adapter.driver.observer_register(self)

        # If the peer (peripheral device) initiates the MTU exchange, this value will be automatically used
        # as this device (Python/interface device) preferred MTU.
        # By default, the Heart Rate Collector example also uses this value to configure the
        # SoftDevice/interface device. See further notes in open()
        self.adapter.default_mtu = 111

        self.mtu_exchanged = False
        self.nus_tx_uuid = None
        self.peer_mtu = 0

    def open(self):
        self.adapter.driver.open()
        if config.__conn_ic_id__.upper() == "NRF51":
            raise AssertionError ("NOT SUPPORTED: config.__conn_ic_id__.upper() == ""NRF51""")
        elif config.__conn_ic_id__.upper() == "NRF52":
            gatt_cfg = BLEConfigConnGatt()

            # In this configuration, we are configuring the MAXIMUM value that the SoftDevice/interface device can
            # ever support. It can only be reconfigured after driver.ble_disable().
            # See: https://infocenter.nordicsemi.com/topic/com.nordic.infocenter.s140.api.v7.2.0/structble__gatt__conn__cfg__t.html#a3fac0ed97a3a75cf21e9ed8432b95193
            gatt_cfg.att_mtu = self.adapter.default_mtu
            gatt_cfg.tag = CFG_TAG
            self.adapter.driver.ble_cfg_set(BLEConfig.conn_gatt, gatt_cfg)

            self.adapter.driver.ble_enable()

    def close(self):
        self.adapter.driver.close()

    def connect_and_discover(self):
        scan_duration = 5
        params = BLEGapScanParams(interval_ms=200, window_ms=150, timeout_s=scan_duration)

        self.adapter.driver.ble_gap_scan_start(scan_params=params)

        try:
            new_conn = self.conn_q.get(timeout=scan_duration)
            self.adapter.service_discovery(new_conn)

            for s in self.adapter.db_conns[new_conn].services:
                # TODO: This is technically an unsafe way to find the NUS Service and Characteristics,
                #       because NUS is a vendor service and we need to check the entire 128 bit UUID
                #       However, at the moment, I have no idea how to get the 128-bit UUID object
                #       for enable_notification().
                if s.uuid.value == 0x0001:

                    for ch in s.chars:
                        # print("ch uuid: {}".format(ch.uuid))
                        # print("ch uuid: {}".format(ch.uuid.__getstate__()))
                        # print("ch uuid: {}".format(ch.uuid.base))
                        if ch.uuid.value == 0x0003:
                            self.nus_tx_uuid = ch.uuid
                            print("Found TX Characteristic. Enabling notification")
                            # self.adapter.enable_notification(new_conn, ch.uuid)

            return new_conn
        except Empty:
            print(f"No NUS device advertising with name {TARGET_DEV_NAME} found.")
            return None

    def on_gattc_evt_prim_srvc_disc_rsp(
        self, ble_driver, conn_handle, status, services
    ):
        services_str = "\n ".join(str(s) for s in services)
        print(
            "on_gattc_evt_prim_srvc_disc_rsp status({}) conn({})\n {}".format(
                status, conn_handle, services_str
            )
        )

    def on_gap_evt_connected(
        self, ble_driver, conn_handle, peer_addr, role, conn_params
    ):
        print("New connection: {}".format(conn_handle))
        self.conn_q.put(conn_handle)

    def on_gap_evt_disconnected(self, ble_driver, conn_handle, reason):
        print("Disconnected: {} {}".format(conn_handle, reason))

    def on_gap_evt_adv_report(
        self, ble_driver, conn_handle, peer_addr, rssi, adv_type, adv_data
    ):
        if BLEAdvData.Types.complete_local_name in adv_data.records:
            dev_name_list = adv_data.records[BLEAdvData.Types.complete_local_name]

        elif BLEAdvData.Types.short_local_name in adv_data.records:
            dev_name_list = adv_data.records[BLEAdvData.Types.short_local_name]

        else:
            return

        dev_name = "".join(chr(e) for e in dev_name_list)
        address_string = "".join("{0:02X}".format(b) for b in peer_addr.addr)
        print(
            "Received advertisement report, address: 0x{}, device_name: {}".format(
                address_string, dev_name
            )
        )

        if dev_name == TARGET_DEV_NAME:
            self.adapter.connect(peer_addr, tag=CFG_TAG)

    def on_notification(self, ble_adapter, conn_handle, uuid, data):
        if uuid == self.nus_tx_uuid:
            print("Connection: {}, {} = {}".format(conn_handle, uuid, data))
            print("{}".format(''.join(chr(c) for c in data)))

    def on_gatts_evt_exchange_mtu_request(self, ble_driver, conn_handle, client_mtu):
        print("abc on_gatts_evt_exchange_mtu_request conn_handle({}) client_mtu({})".format(conn_handle, client_mtu))
        self.mtu_exchanged = True
        self.peer_mtu = client_mtu

    def on_gattc_evt_exchange_mtu_rsp(self, ble_driver, conn_handle, status, att_mtu):
        print("abc on_gattc_evt_exchange_mtu_rsp conn_handle({}) status ({}) att_mtu({})".format(conn_handle, status, att_mtu))
        self.mtu_exchanged = True
        self.peer_mtu = att_mtu


def main(selected_serial_port):
    print("Serial port used: {}".format(selected_serial_port))
    driver = BLEDriver(
        serial_port=selected_serial_port, auto_flash=False, baud_rate=1000000, log_severity_level="info"
    )

    adapter = BLEAdapter(driver)
    nus_client = NusClient(adapter)
    nus_client.open()
    conn = nus_client.connect_and_discover()

    if conn is not None:

        if not nus_client.mtu_exchanged:
            print("MTU Exchange will now be initiated by the Python Central")

            try:
                requested_mtu = adapter.default_mtu
                adapter.driver.ble_gattc_exchange_mtu_req(conn_handle=conn, mtu=requested_mtu)

                while not nus_client.mtu_exchanged:
                    time.sleep(1)

                if nus_client.peer_mtu > requested_mtu:
                    print("Final MTU: {}".format(requested_mtu))
                else:
                    print("Final MTU: {}".format(nus_client.peer_mtu))

            except NordicSemiException as ex:
                logger.exception("NordicSemiException ble_gattc_exchange_mtu_req: {}".format(ex))
            except Exception as ex:
                logger.exception("Exception ble_gattc_exchange_mtu_req: {}".format(ex))
        else:
            print("MTU Exchange was initiated by the Peripheral peer")
            if nus_client.peer_mtu > adapter.default_mtu:
                print("Final MTU: {}".format(adapter.default_mtu))
            else:
                print("Final MTU: {}".format(nus_client.peer_mtu))

        time.sleep(5)
        adapter.disconnect(conn)
    nus_client.close()

    while True:
        time.sleep(600)


def item_choose(item_list):
    for i, it in enumerate(item_list):
        print("\t{} : {}".format(i, it))
    print(" ")

    while True:
        try:
            choice = int(input("Enter your choice: "))
            if (choice >= 0) and (choice < len(item_list)):
                break
        except Exception:
            pass
        print("\tTry again...")
    return choice


if __name__ == "__main__":
    logging.basicConfig(
        level="DEBUG",
        format="%(asctime)s [%(thread)d/%(threadName)s] %(message)s",
    )
    serial_port = None
    if len(sys.argv) < 2:
        print("Please specify connectivity IC identifier (NRF51, NRF52)")
        exit(1)
    init(sys.argv[1])
    if len(sys.argv) == 3:
        serial_port = sys.argv[2]
    else:
        descs = BLEDriver.enum_serial_ports()
        choices = ["{}: {}".format(d.port, d.serial_number) for d in descs]
        choice = item_choose(choices)
        serial_port = descs[choice].port
    main(serial_port)
    quit()
